

from baopig.lib import HoverableByMouse
from .text import Text, DynamicText


def _init_loc(location):
    """ location: 'top', 'bottom', 'left' or 'right' """

    assert location in ("top", "bottom", "left", "right")
    if location == "top":
        loc_opposite = "midbottom"
        pos = (0, -5)
    elif location == "bottom":
        loc_opposite = "midtop"
        pos = (0, 5)
    elif location == "left":
        loc_opposite = "midright"
        pos = (-5, 0)
    elif location == "right":
        loc_opposite = "midleft"
        pos = (5, 0)
    else:
        raise ValueError(f"location must be 'top', 'bottom', 'left' or 'right'. Wrong value : {location}")

    return pos, loc_opposite


class Indicator(Text):

    def __init__(self, widget, text, indicator=None, loc="top", **kwargs):
        """Create a Text above the widget when hovered"""

        assert isinstance(widget, HoverableByMouse), "Indicator can only indicate HoverableByMouse widgets"
        if widget._indicator is not None:
            raise PermissionError("A Widget can only have one indicator")
        if indicator is not None:
            widget._indicator = indicator
            raise NotImplementedError

        pos, loc_opposite = _init_loc(loc)
        loc = "mid" + loc
        kwargs["loc"] = loc_opposite

        Text.__init__(
            self, widget.parent, text, font_color=(255, 255, 255), font_height=15, pos=pos,
            ref=widget, refloc=loc, referenced_by_hitbox=True,
            background_color=(0, 0, 0, 192), padding=(8, 4), selectable=False, layer_level=2, **kwargs
        )

        widget._indicator = self
        widget.signal.HOVER.connect(self.wake, owner=self)
        widget.signal.UNHOVER.connect(self.sleep, owner=self)
        if not widget.is_hovered:
            self.sleep()


class DynamicIndicator(DynamicText):

    def __init__(self, widget, get_text, indicator=None, loc="top", **kwargs):
        """Create a DynamicText above the widget when hovered"""

        if widget._indicator is not None:
            raise PermissionError("A Widget can only have one indicator")
        if indicator is not None:
            widget._indicator = indicator
            raise NotImplementedError

        pos, loc_opposite = _init_loc(loc)
        loc = "mid" + loc
        kwargs["loc"] = loc_opposite

        DynamicText.__init__(
            self, widget.parent, get_text, font_color=(255, 255, 255), font_height=15, pos=pos,
            ref=widget, refloc=loc, referenced_by_hitbox=True,
            background_color=(0, 0, 0, 192), padding=(8, 4), selectable=False, layer_level=2, **kwargs
        )

        widget._indicator = self
        widget.signal.HOVER.connect(self.wake, owner=self)
        widget.signal.UNHOVER.connect(self.sleep, owner=self)
        if not widget.is_hovered:
            self.sleep()





# TODO : move to prefabs folder
from baopig import *


class TesterScene(Scene):

    def __init__(self, app, content_zone_class):

        Scene.__init__(self, app, name=str(content_zone_class))
        self.sections = []

        # Menu
        self.menu_zone = Zone(self, size=("50%", 35), pos=(0, 10), name="menu")

        # Section
        self.sections_zone = Zone(self, size=("50%", self.rect.h - self.menu_zone.rect.h - 20),
                                  background_color="lightgray",
                                  pos=(0, self.menu_zone.rect.bottom + 10), name="sections")
        GridLayer(self.sections_zone, name="sections_layer", col_width=self.sections_zone.rect.w, nbcols=1)
        self.add_section(title="NO TEST SECTION YET", tests=[])

        # Run button
        def run():
            import tempfile
            with tempfile.NamedTemporaryFile(suffix=".py", mode="w", delete=False) as script:
                script.write(self.try_it_yourself.code.text)
                script.seek(0)
                script.close()

                import subprocess
                with tempfile.TemporaryFile(mode="r") as output_file:
                    subprocess.call(f"python3 {script.name}", shell=True, stdout=output_file, stderr=output_file)
                    output_file.seek(0)
                    output = output_file.read()
                    self.try_it_yourself.console.set_text(str(output))
        self.run_tiy = Button(self, "RUN", command=run, sticky="topright", ref=self.menu_zone, visible=False)

        # Back & Try_it_yourself buttons
        GridLayer(self.menu_zone, nbrows=1, spacing=10)
        Button(self.menu_zone, "< BACK", col=0, command=PrefilledFunction(app.open, "UTMenu_Scene"))

        def try_it_yourself():
            if self.try_it_yourself is None:
                self._init_try_it_yourself()

            if self.try_it_yourself.is_visible:
                self.try_it_yourself.hide()
                self.run_tiy.hide()
                self.sections_zone.show()
            else:
                self.try_it_yourself.show()
                self.run_tiy.show()
                self.sections_zone.hide()

        Button(self.menu_zone, "Try it yourself !", col=1, command=try_it_yourself)

        # Content
        self.content = content_zone_class(self, size=("50%", self.rect.h - 20), name="content", pos=(0, 10),
                                          sticky="topright")
        if hasattr(self.content, "load_sections"):
            self.content.load_sections()

        # Try it yourself
        self.try_it_yourself = None

    def _init_try_it_yourself(self):

        import inspect
        file = open(inspect.getfile(self.content.__class__), "r")
        content = file.read()
        file.close()

        content = content.split("\n\n# For the PresentationScene import")

        code = f"""# Default code for Try it yourself
{content[0]}

app = Application(size=(700, 700))
main_zone = {self.content.__class__.__name__}(Scene(app), size=("90%", "90%"), sticky="center")
app.launch()"""

        self.try_it_yourself = Zone(self, size=("50%", self.rect.h - 20), spacing=10,
                                    pos=(0, self.menu_zone.rect.bottom + 10), name="try_it_yourself")
        code_window = Zone(self.try_it_yourself, size=("100%", 550))
        self.try_it_yourself.code = TextEdit(code_window, text=code, font_file="monospace", width="100%", height=550,
                                             padding=(2, 0, 30, 60))
        self.try_it_yourself.console = Text(self.try_it_yourself, font_file="monospace",
                                            background_color=(211, 189, 189))

        self.try_it_yourself.pack()
        self.try_it_yourself.hide()

    def add_section(self, title, tests):

        if self.sections and self.sections[0] == ["NO TEST SECTION YET", []]:
            self.sections.pop(0)
            for temp in tuple(self.sections_zone.children):
                temp.kill()

        self.sections.append([title, tests])
        Text(self.sections_zone, f"--- SECTION {len(self.sections)} : {title} ---",
             font_height=self.theme.get_style_for(Text)["font_height"] + 2, font_bold=True,
             row=len(self.sections_zone.children), max_width=self.sections_zone.rect.w)
        for i, v in enumerate(tests):
            cb = CheckBox(self.sections_zone, text=f"TEST {i + 1} : {v}",  # {:0>2} for 01, 02...
                          row=len(self.sections_zone.children), width=self.sections_zone.rect.w)
            if v.startswith("TODO"):
                cb.text_widget.font.config(color="red3")
        Text(self.sections_zone, "", row=len(self.sections_zone.children))

    def handle_scene_close(self):

        self.kill()



from baopig import *


class UT_Sleep_Zone(Zone):
    def __init__(self, *args, **kwargs):
        Zone.__init__(self, *args, **kwargs)

        Layer(self, name="zones_layer", spacing=10)
        z1 = Zone(self, size=("100%", 50), background_color=(150, 150, 150), padding=10, spacing=5)
        z2 = Zone(self, size=("100%", 250), background_color=(150, 150, 150))
        z3 = Zone(self, size=("100%", 335), background_color=(150, 150, 150))
        self.default_layer.pack()

        # Z1
        def tog(widg):
            clicked_button = mouse.hovered_widget
            if widg.is_dead:
                clicked_button.text_widget.set_text("state:DEAD")
            elif widg.is_asleep:
                widg.wake()
                if widg.is_dead:
                    clicked_button.text_widget.set_text("state:DEAD")
                else:
                    clicked_button.text_widget.set_text("state:AWAKE")
            else:
                if widg.is_alive:
                    widg.sleep()
                    clicked_button.text_widget.set_text("state:ASLEEP")

        main = Text(z1, text="sleepy Text", font_color="blue", background_color="blue4", padding=5, size=(100, 30))
        Button(z1, text="state:AWAKE", command=PrefilledFunction(tog, main), height=30)
        r2 = Rectangle(z1, color="red")
        main.signal.SLEEP.connect(lambda: r2.set_color("black"), owner=r2)
        main.signal.WAKE.connect(lambda: r2.set_color("red3"), owner=r2)
        z1.default_layer.pack(axis="horizontal")

        # Z2
        def add():
            if z2.test_zone.texts.is_dead:
                z2.test_zone.texts = Zone(z2.test_zone, size=(z2.rect.w - 100, "100%"))
                z2.test_zone.togglers.default_layer.clear()
                z2.test_zone.default_layer.pack(axis="horizontal")
            t = Text(z2.test_zone.texts, text=f"index:{z2.index}",
                     width="100%", border_width=2, padding=5, align_mode="center", pos=(0, z2.index * 35))
            Button(z2.test_zone.togglers, text="state:AWAKE", command=PrefilledFunction(tog, t), pos=t.rect.topleft)
            z2.index += 1

        z2.index = 0
        z2.param_zone = Zone(z2, size=("100%", 35))
        z2.test_zone = Zone(z2, size=("100%", z2.rect.h - z2.param_zone.rect.h), background_color=(130, 130, 130))
        z2.default_layer.pack()
        z2.test_zone.texts = Zone(z2.test_zone, size=(z2.rect.w - 100, "100%"))
        z2.test_zone.togglers = Zone(z2.test_zone, size=(100, "100%"))
        z2.test_zone.default_layer.pack(axis="horizontal")
        b = Button(z2.param_zone, "ADD", command=add)
        DynamicText(z2.param_zone, get_text=lambda: f"LEN:{len(z2.test_zone.texts.children)}",
                    size=b.rect.size, align_mode="center", padding=(0, b.text_widget.rect.top))

        def clear():
            if z2.test_zone.texts.is_alive:
                z2.test_zone.texts.default_layer.clear()
            z2.test_zone.togglers.default_layer.clear()
            z2.index = 0

        Button(z2.param_zone, "CLEAR", command=clear)

        def kill():
            if z2.test_zone.texts.is_alive:
                z2.test_zone.texts.kill()
                z2.index = 0

        Button(z2.param_zone, "KILL", command=kill)
        z2.param_zone.default_layer.pack(axis="horizontal")

        # Z3
        class DraggableRectangle(Rectangle, DraggableByMouse):
            def __init__(self, parent, **kwargs2):
                Rectangle.__init__(self, parent, **kwargs2)
                DraggableByMouse.__init__(self, parent, **kwargs2)

        z3.set_style_for(Rectangle, color=(110, 80, 90))
        clone_zone = Zone(z3, size=("100%", 150))
        original_zone = Zone(z3, size=("100%", 150))
        original = DraggableRectangle(original_zone, color=(116, 0, 32))
        clone = Rectangle(clone_zone, size=original.rect.size, pos=(0, - 150 - 35), ref=original)
        clone2 = Rectangle(clone_zone, size=clone.rect.size, pos=(50, 0), ref=clone, color=(110, 100, 100))
        # TODO : solve: an adaptable size is not linked to the ref but to the parent
        b = Button(z3, text="state:AWAKE", width="25%", command=PrefilledFunction(tog, clone))
        b.move_behind(original_zone)
        z3.default_layer.pack()

        def handle_resize():
            clone.resize(*original.rect.size)

        original.signal.RESIZE.connect(handle_resize, owner=clone)

        def handle_new_surface():
            cloned_color = original.color.copy()
            cloned_color.s -= 50
            clone.set_color(cloned_color)

        original.signal.NEW_SURFACE.connect(handle_new_surface, owner=clone)
        handle_new_surface()

        def handle_resize():
            clone2.resize(*clone.rect.size)

        clone.signal.RESIZE.connect(handle_resize, owner=clone2)

        def handle_new_surface():
            cloned_color = clone.color.copy()
            cloned_color.s -= 10
            clone2.set_color(cloned_color)

        clone.signal.NEW_SURFACE.connect(handle_new_surface, owner=clone2)
        handle_new_surface()

        def resize():
            if original.rect.width == 30:
                original.resize(40, 40)
            elif original.rect.width == 40:
                original.resize(20, 20)
            else:
                original.resize(30, 30)

        b2 = Button(z3, text="Resize", width="25%", ref=b, refloc="topright", command=resize)

        def toggle_visibility():
            clicked_button = mouse.hovered_widget
            if clone.is_visible:
                clone.hide()
                clicked_button.text_widget.set_text("state:HIDDEN")
            else:
                clone.show()
                clicked_button.text_widget.set_text("state:VISIBLE")

        b3 = Button(z3, text="state:VISIBLE", width="25%", ref=b2, refloc="topright", command=toggle_visibility)

        def toggle_color():
            clicked_button = mouse.hovered_widget
            text = clicked_button.text_widget.text
            if text == "state:RED":
                clicked_button.text_widget.set_text("state:GREEN")
            elif text == "state:GREEN":
                clicked_button.text_widget.set_text("state:BLUE")
            else:
                clicked_button.text_widget.set_text("state:RED")
            original.color.h = (original.color.h + 120) % 360
            original.send_paint_request()

        Button(z3, text="state:RED", width="25%", ref=b3, refloc="topright", command=toggle_color)

        # Z5
        # TODO : try to asleep a widget from a BoxLayout, and then wake it up

    def load_sections(self):
        """
        If the parent got killed:
            Kills the widget
        Attaches the widget to its parent
        Updates the widget's size & position
        Emits the WAKE signal
        """

        self.parent.add_section(
            title="Sleeping state",
            tests=[
                "When a sleeping widget wakes up, the signal WAKE is emitted",
                "When a widget goes to sleep, the signal SLEEP is emitted",
                "A sleeping widget is not inside its parent",
                "When a sleeping widget wakes up, if its parent is dead, the widget dies",
                "When a sleeping widget wakes up, if its parent got resized, the widget reacts",
                "When a sleeping widget wakes up, if its parent moved, the widget reacts",
                "Widget.hide() & Widget.show() are alright with sleepy widgets",
            ]
        )


# For the PresentationScene import
ut_zone_class = UT_Sleep_Zone

if __name__ == "__main__":
    from baopig.prefabs.testerscene import TesterScene
    app = Application()
    TesterScene(app, ut_zone_class)
    app.launch()





class BaopigVersion(tuple):
    """Code from pygame.version.PygameVersion"""
    __slots__ = ()
    fields = 'major', 'minor', 'patch', 'backup'

    def __new__(cls, major, minor, patch, backup=None):
        return tuple.__new__(cls, (major, minor, patch) + tuple([backup] if backup else ()))

    def __repr__(self):
        fields = ('{}={}'.format(fld, val) for fld, val in zip(self.fields, self))
        return '{}({})'.format(str(self.__class__.__name__), ', '.join(fields))

    def __str__(self):
        return '{}.{}.{}'.format(*self) if len(self) == 3 else '{}.{}.{} - {}'.format(*self)

    major = property(lambda self: self[0])
    minor = property(lambda self: self[1])
    patch = property(lambda self: self[2])


version = BaopigVersion(0, 19, 0)

if __name__ == "__main__":
    # TODO : try to access baopig.version without creating application, threads...
    print("Hello, this is baopig version", version)




from baopig.io import mouse
from baopig.lib import Rectangle, Container, LinkableByMouse
from .indicator import DynamicIndicator
from .text import Text


class SliderBloc(Rectangle):
    STYLE = Rectangle.STYLE.substyle()
    STYLE.modify(
        width=0,  # use length and wideness instead
        height=0,  # use length and wideness instead
        border_width=3,
    )
    STYLE.create(
        length=16,
        wideness="100%",
    )
    STYLE.set_constraint("height", lambda val: val == 0, "Use length and wideness instead of width and height")
    STYLE.set_constraint("width", lambda val: val == 0, "Use length and wideness instead of width and height")

    def __init__(self, slider):
        assert isinstance(slider, Slider)

        sticky = "midleft" if slider.axis == 'x' else "midtop"
        Rectangle.__init__(self, slider, sticky=sticky)

        assert self.rect.size == (0, 0), "Use length and wideness instead of size, width and height"

        self._length = self.style["length"]
        self._wideness = self.style["wideness"]
        self._max_x = None

        self.signal.RESIZE.connect(self.handle_resize, owner=None)

        if slider.axis == "x":
            self.resize(width=self._length, height=self._wideness)
        else:
            self.resize(width=self._wideness, height=self._length)

    slider = property(lambda self: self._parent)
    max_x = property(lambda self: self._max_x)

    def update(self):
        if self.slider.axis == "x":
            self.set_pos(left=self.slider.get_pourcent() * self.max_x)
        else:
            self.set_pos(top=self.slider.get_pourcent() * self.max_x)

    def handle_resize(self):

        if self.slider.axis == "x":
            self._max_x = self.slider.rect.width - self.rect.width
        else:
            self._max_x = self.slider.rect.height - self.rect.height


class SliderBar(Rectangle):
    STYLE = Rectangle.STYLE.substyle()
    STYLE.modify(
        loc="center",
        refloc="center",
        width=0,  # use length and wideness instead
        height=0,  # use length and wideness instead
        color=(0, 0, 0, 64),
        border_width=1,
    )
    STYLE.create(
        wideness="90%",  # length is defined by the slider
    )
    STYLE.set_constraint("height", lambda val: val == 0, "Use length and wideness instead of width and height")
    STYLE.set_constraint("width", lambda val: val == 0, "Use length and wideness instead of width and height")

    # NOTE : we replaced width/height with length/wideness, so it is easier to code vertical sliders

    def __init__(self, slider):
        assert isinstance(slider, Slider)

        # style = slider.get_style_for(self.__class__)
        # self.inherit_style_(slider)
        # TODO : if the slider goes vertically, switch length and wideness in the following line

        Rectangle.__init__(self, slider)

        assert self.rect.size == (0, 0), "Use length and wideness instead of size, width and height"

        self._length = 0  # self.style["length"]  TODO : remove length, defined by slider
        self._wideness = self.style["wideness"]
        if slider.axis == "x":
            self.resize(width=self._length, height=self._wideness)
        else:
            self.resize(width=self._wideness, height=self._length)


class Slider(Container, LinkableByMouse):
    """Widget that contains a bar and a slideable bloc"""

    STYLE = Container.STYLE.substyle()
    STYLE.modify(
        width=0,  # don't use them
        height=0,  # don't use them
    )
    # NOTE : On peut facilement se tromper en laissant width et height alors qu'on devrait utiliser bar_size
    STYLE.create(
        length=150,
        wideness=15,
        has_indicator=True,
        bloc_class=SliderBloc,
        bar_class=SliderBar,
        axis="x",
    )
    STYLE.set_type("has_indicator", bool)
    STYLE.set_constraint("axis", lambda val: val in ("x", "y"), "must be 'x' or 'y'")
    STYLE.set_constraint("bloc_class", lambda val: issubclass(val, SliderBloc))
    STYLE.set_constraint("bar_class", lambda val: issubclass(val, SliderBar))
    STYLE.set_constraint("height", lambda val: val == 0, "Use length and wideness instead of width and height")
    STYLE.set_constraint("width", lambda val: val == 0, "Use length and wideness instead of width and height")

    def __init__(self, parent, minval, maxval, defaultval=None, step=None, title=None, printed_title=False, **kwargs):

        if defaultval is None:
            defaultval = minval

        assert minval <= maxval, f"There must be a positive difference between minval and maxval " \
                                 f"(minval : {minval}, maxval : {maxval})"
        assert minval <= defaultval <= maxval, f"The defaultval must be included between minval and maxval " \
                                               f"(minval : {minval}, maxval : {maxval}, defaultval : {defaultval})"
        if step is not None:
            assert step > 0

        Container.__init__(self, parent, **kwargs)
        LinkableByMouse.__init__(self, parent)

        self._minval = minval
        self._maxval = maxval
        self._range = self.maxval - self.minval
        self._defaultval = self._val = defaultval
        self._step = step
        self._link_origin = None  # the link's position, relative to self
        self._axis = self.style["axis"]
        self._length = self.style["length"]
        self._wideness = self.style["wideness"]

        self.create_signal("NEW_VAL")

        self.bar = self.style["bar_class"](self)
        self.bloc = self.style["bloc_class"](self)

        assert self.rect.size == (0, 0), "Use length and wideness instead of size, width and height"
        if self.axis == "x":
            self.resize(width=self._length, height=self._wideness)
        else:
            self.resize(width=self._wideness, height=self._length)

        self.bloc.update()

        if self.style["has_indicator"]:
            if title:
                if printed_title:
                    self.title = Text(self, title, sticky="center", selectable=False, font_color=(96, 96, 96),
                                      font_height=int((self.bar.rect.height - self.bar.border_width * 2) * .9),
                                      font_bold=True)
                    self.title.set_touchable_by_mouse(False)
                DynamicIndicator(self, get_text=lambda: f"{title} : {self.val}")
            else:
                DynamicIndicator(self, get_text=lambda: self.val)

    axis = property(lambda self: self._axis)
    defaultval = property(lambda self: self._defaultval)
    maxval = property(lambda self: self._maxval)
    minval = property(lambda self: self._minval)
    range = property(lambda self: self._range)
    step = property(lambda self: self._step)
    val = property(lambda self: self._val)

    def _update_val(self, val=None, x=None):

        assert (val is None) != (x is None)

        if x is not None:
            current_x = self.bloc.rect.left if self.axis == "x" else self.bloc.rect.top
            if x == current_x:
                return
            # val = x * (max - min) / max_index + min
            val = x * self.range / self.bloc.max_x + self.minval
        if self.step is not None:
            def cut(n, length):
                # print(n, l, float(("{:." + str(l-1) + "e}").format(n)))
                return float(("{:." + str(length - 1) + "e}").format(n))

            val = round((val - self.minval) / self.step) * self.step + self.minval
            if isinstance(self.step, float):
                val = cut(val, len(str(self.step % 1)) - 2 + len(str(int(val))))
            if isinstance(val, float) and val.is_integer():  # remove .0 for the beauty of the indicator
                val = int(val)
            if val >= self.maxval:  # not else, because step can make val go to maxval or higher
                val = self.maxval

        if val == self.val:
            return
        assert self.minval <= val <= self.maxval, f"{val}, {self.maxval}, {self.minval}"

        self._val = val
        # x = (val - min) / (max - min) * max_index
        self.bloc.update()
        current_x = self.bloc.rect.left if self.axis == "x" else self.bloc.rect.top
        if current_x == 0:
            self._val = self.minval  # prevent approximations
        self.signal.NEW_VAL.emit(self.val)

    def get_pourcent(self):
        """Return the percentage from min to val in the range min -> max"""
        if self.range:
            return (self.val - self.minval) / self.range
        else:
            return 0

    def handle_link(self):

        def clamp(val, min_, max_):
            """Clamp f between min and max"""
            return min(max(val, min_), max_)

        size_index = 0 if self.axis == "x" else 1
        if self.bloc.collidemouse():
            self._link_origin = mouse.get_pos_relative_to(self.bloc)[size_index]
        else:
            size_val = self.bloc.rect.width if self.axis == "x" else self.bloc.rect.height
            self._link_origin = size_val / 2
            self._update_val(x=clamp(mouse.get_pos_relative_to(self)[size_index] - self._link_origin,
                                     0, self.bloc.max_x))

    def handle_link_motion(self, rel):

        def clamp(val, min_, max_):
            """Clamp f between min and max"""
            return min(max(val, min_), max_)

        size_index = 0 if self.axis == "x" else 1
        x = clamp(
            mouse.get_pos_relative_to(self)[size_index] - self._link_origin,
            0, self.bloc.max_x
        )
        self._update_val(x=x)

    def handle_resize(self):

        super().handle_resize()
        # TODO : bloc.length, bloc.wideness
        bar_margin = max(0, self.bloc.border_width - self.bar.border_width) * 2
        if self.axis == "x":
            self.bar.resize_width(self.rect.width - bar_margin)
        else:
            self.bar.resize_height(self.rect.height - bar_margin)

    def resize_TBR(self, w, h):
        pass
        # TODO : bloc.length, bloc.wideness

    def reset(self):
        """Set the value to defaultval"""
        if self.val == self.defaultval:
            return
        self._update_val(self.defaultval)

    def set_defaultval(self, val, reset=True):
        """If reset is True, reset the value to defaultval"""

        if val is self.defaultval:
            return
        assert self.minval <= val <= self.maxval, f"The value must be included between minval and maxval " \
                                                  f"(minval : {self.minval}, maxval : {self.maxval}, startval : {val})"

        self._defaultval = val
        if reset:
            self.reset()



from baopig.lib import *
from .button import Button
from .text import Text


class DialogAnswerButton(Button):

    dialog = property(lambda self: self.scene)

    def handle_validate(self):

        self.dialog._answer(self.text_widget.text)


class DialogButtonsZone(Zone):
    STYLE = Zone.STYLE.substyle()
    STYLE.modify(
        pos=(0, -30),
        loc="midbottom",
        refloc="midbottom",
    )

    def __init__(self, dialog_frame):

        assert isinstance(dialog_frame, DialogFrame)

        choices = dialog_frame.parent.choices
        Zone.__init__(
            self, dialog_frame,
            size=(dialog_frame.rect.w - 60, int(46 * ((len(choices) - 1) / 3 + 1))),
        )
        grid = GridLayer(self, nbrows=int((len(choices) - 1) / 3) + 1, nbcols=min(len(choices), 3),
                         row_height=46, spacing=10)
        grid.set_col_width(int((self.rect.w - (grid.nbcols - 1) * grid.spacing.left) / grid.nbcols))
        for i, choice in enumerate(choices):
            assert isinstance(choice, str), "Other types are not implemented"
            self.dialog.style["answerbutton_class"](self, choice, col=i % 3, row=i // 3, loc="center")

    def get_dialog(self):
        dialog = self.parent
        while not isinstance(dialog, Dialog):
            if isinstance(dialog, Scene):
                raise TypeError("A DialogAnswerButton must be inside a Dialog")
            dialog = dialog.parent
        return dialog

    dialog = property(get_dialog)


class DialogFrame(Zone):
    STYLE = Zone.STYLE.substyle()
    STYLE.modify(
        pos=("50%", "50%"),
        loc="center",
        width=450,
        height=300,
        background_color="theme-color-scene_background",
    )

    def __init__(self, dialog, **kwargs):
        assert isinstance(dialog, Dialog)

        Zone.__init__(self, dialog, **kwargs)

        self.buttons_zone = dialog.style["buttonszone_class"](self)

        self.title_label = Text(
            self, dialog.title,
            font_height=38,
            center=("50%", 40),
        )
        bottom = self.title_label.rect.bottom
        if dialog.description is not None:
            self.description_label = Text(
                self, dialog.description,
                font_height=27, max_width=self.rect.w - 60,
                pos=(30, self.title_label.rect.bottom + 15),
            )
            bottom = self.description_label.rect.bottom

        self.resize_height(max(bottom + 50 + self.buttons_zone.rect.height + 10, self.rect.height))


class Dialog(Scene):
    """
    Example :
        dialog = bp.Dialog(app, "Votre adversaire a quitté la partie en cours", choices=("OK",))
        def click_end_dialog(choice):
            app.open("menu")
        dialog.signal.ANSWERED.connect(click_end_dialog)

    If one_shot is True, this dialog will kill itself after the first answer
    """

    STYLE = Scene.STYLE.substyle()
    STYLE.create(
        title="Dialog",  # TODO : why are these style attributes ??
        description=None,
        choices=("Cancel", "Continue"),
        default_choice_index=0,  # focuses the first answer button
        dialogframe_class=DialogFrame,
        buttonszone_class=DialogButtonsZone,
        answerbutton_class=DialogAnswerButton,
    )
    STYLE.set_type("title", str)
    STYLE.set_type("default_choice_index", int)
    STYLE.set_constraint("dialogframe_class", lambda val: issubclass(val, DialogFrame),
                         "must be a subclass of DialogFrame")
    STYLE.set_constraint("buttonszone_class", lambda val: issubclass(val, DialogButtonsZone),
                         "must be a subclass of DialogButtonsZone")
    STYLE.set_constraint("answerbutton_class", lambda val: issubclass(val, DialogAnswerButton),
                         "must be a subclass of DialogAnswerButton")

    def __init__(self, app, one_shot=False, **kwargs):

        Scene.__init__(self, app, **kwargs)

        self.title = self.style["title"]
        self.choices = self.style["choices"]
        self.description = self.style["description"]
        self.default_choice_index = self.style["default_choice_index"]
        assert self.default_choice_index in range(len(self.choices))
        self.frame = self.style["dialogframe_class"](self)  # TODO : a class style cannot be a parameter

        self.hovered_scene = None
        self.answer = None
        self.create_signal("ANSWERED")
        self.one_shot = one_shot

    def _answer(self, ans):
        """Only called by DialogAnswerButton"""
        self.answer = ans
        self.hovered_scene.open()
        self.signal.ANSWERED.emit(self.answer)
        if self.one_shot:
            self.kill()

    def handle_scene_open(self):
        self.focus(self.frame.buttons_zone.default_layer[self.default_choice_index])

    def open(self):

        if self.application.focused_scene is self:
            return

        app = self.application
        self.hovered_scene = app.focused_scene
        background_image = app.focused_scene.surface.copy()
        sail = pygame.Surface(app.size, pygame.SRCALPHA)
        sail.fill((0, 0, 0, 100))
        background_image.blit(sail, (0, 0))
        self.set_background_image(background_image)

        super().open()



import pygame
from baopig.pybao.objectutilities import Object
from baopig.lib import Focusable, Layer, Validable
from baopig.lib import Rectangle, Image, Container
from .text import Text


class AButton_FocusSail(Rectangle):
    STYLE = Rectangle.STYLE.substyle()
    STYLE.modify(width="100%", height="100%", color=(0, 0, 0, 0), border_color="theme-color-border", border_width=1)

    def __init__(self, abutton):
        Rectangle.__init__(self, abutton, visible=False, layer=abutton.behind_content)


class AButton_HoverSail(Rectangle):
    STYLE = Rectangle.STYLE.substyle()
    STYLE.modify(width="100%", height="100%")
    STYLE.create(alpha=63)

    def __init__(self, abutton):
        Rectangle.__init__(self, abutton, visible=False, layer=abutton.above_content)
        self.set_color((0, 0, 0, self.style["alpha"]))


class AButton_LinkSail(Rectangle):
    STYLE = Rectangle.STYLE.substyle()
    STYLE.modify(width="100%", height="100%")
    STYLE.create(alpha=63)

    def __init__(self, abutton):
        Rectangle.__init__(self, abutton, visible=False, layer=abutton.behind_content)
        self.set_color((0, 0, 0, self.style["alpha"]))


class AbstractButton(Container, Focusable, Validable):
    """
    Abstract button

    A button is clicked when the link is released while the mouse is still
    hovering it (for more explanations about links, see LinkableByMouse)

    - background color
    - focus
    - link
    - text
    - hover
    """

    STYLE = Container.STYLE.substyle()
    STYLE.modify(
        width=100,
        height=35,
        background_color="theme-color-content",
        padding=10,
    )
    STYLE.create(
        catching_errors=False,
        focus_class=AButton_FocusSail,
        hover_class=AButton_HoverSail,
        link_class=AButton_LinkSail,
    )

    def __init__(self, parent, command=None, hover=None, link=None, focus=None, **kwargs):

        Container.__init__(self, parent, **kwargs)
        Focusable.__init__(self, parent, **kwargs)
        Validable.__init__(self, catching_errors=self.style["catching_errors"])

        if command is None:
            self.command = lambda: None
        else:
            self.command = command  # non protected field
            assert callable(command), "command must be callable"

        self.behind_content = Layer(self, weight=0)
        self.content = Layer(self, weight=1)
        self.above_content = Layer(self, weight=2)

        void = Object(show=lambda: None, hide=lambda: None)

        self.set_style_for(Rectangle, width="100%", height="100%")
        self.set_style_for(Image, width="100%", height="100%")  # TODO : implement (this is not working)

        hover_class = self.style["hover_class"]
        if hover_class:
            self._hover_sail_ref = hover_class(self).get_weakref()
        else:
            self._hover_sail_ref = lambda: void

        focus_class = self.style["focus_class"]
        if focus_class:
            self._focus_sail_ref = focus_class(self).get_weakref()
        else:
            self._focus_sail_ref = lambda: void

        link_class = self.style["link_class"]
        if link_class:
            self._link_sail_ref = link_class(self).get_weakref()
        else:
            self._link_sail_ref = lambda: void

        self._disable_sail_ref = Rectangle(  # TODO : same as hover, focus and link
            self, color=(255, 255, 255, 128), visible=False, layer=self.above_content, name=self.name + ".disable_sail",
        ).get_weakref()

        if isinstance(hover, int) and hover != -1:  # TODO : hover_alpha=False or Sail
            # Adapts the hover sail to the surface -> alpha pixels are not hovered
            hidden = self.is_hidden
            if hidden:
                if self.has_locked("visibility"):
                    raise NotImplementedError
                self.show()
            self._flip_without_update()
            if hidden:
                self.hide()
            self.hover_sail.surface.blit(self.surface, (0, 0), special_flags=pygame.BLEND_RGBA_MIN)

        self.signal.KILL.connect(self.handle_kill, owner=None)

    disable_sail = property(lambda self: self._disable_sail_ref())
    focus_sail = property(lambda self: self._focus_sail_ref())
    hover_sail = property(lambda self: self._hover_sail_ref())
    link_sail = property(lambda self: self._link_sail_ref())

    def disable(self):

        self.set_touchable_by_mouse(False)

    def enable(self):

        self.set_touchable_by_mouse(True)

    def handle_defocus(self):

        self.focus_sail.hide()

    def handle_focus(self):

        self.focus_sail.show()

    def handle_hover(self):

        self.hover_sail.show()

    def handle_keydown(self, key):

        if key == pygame.K_RETURN:
            self.validate()
        else:
            super().handle_keydown(key)

    def handle_keyup(self, key):

        if key == pygame.K_RETURN:
            self.link_sail.hide()  # For validation via RETURN key

    def handle_kill(self):

        void = Object(show=lambda: None, hide=lambda: None)
        self._link_sail_ref = lambda: void

    def handle_link(self):

        self.link_sail.show()

    def handle_unhover(self):

        self.hover_sail.hide()

    def handle_unlink(self):
        # TODO : test : if a linked Clickable is hidden, handle_unlink() is called but validate() is not called

        if self.collidemouse():
            self.validate()

        self.link_sail.hide()

    def handle_validate(self):

        self.command()
        self.link_sail.show()  # For validation via RETURN key

    def set_touchable_by_mouse(self, val):

        super().set_touchable_by_mouse(val)

        if self.is_touchable_by_mouse:
            self.disable_sail.hide()
        else:
            self.disable_sail.show()


class Button_Text(Text):
    STYLE = Text.STYLE.substyle()
    STYLE.modify(
        align_mode="left",
        loc="center",
        refloc="center",
    )

    def __init__(self, button, text, **kwargs):

        assert isinstance(button, AbstractButton)

        Text.__init__(self, button, text=text, selectable=False, **kwargs)

        assert self.width_is_adaptable

        content_rect = button.content_rect
        if content_rect.height < self.font.height:
            self.font.config(height=content_rect.height)  # changing the font will automatically update the text
        while self.rect.width > content_rect.width or self.rect.height > content_rect.height:
            if self.font.height == 2:
                break
                # raise ValueError(f"This text is too long for the text area : {text} (area={content_rect})")
            self.font.config(height=self.font.height - 1)  # changing the font will automatically update the text
        # if self.rect.height > content_rect.height:
        #     self.resize_height(content_rect.height)


class Button(AbstractButton):
    """
    Un Button est un bouton classique, avec un text
    Sa couleur d'arriere plan par defaut est (64, 64, 64)
    """

    STYLE = AbstractButton.STYLE.substyle()
    STYLE.create(
        text_class=Button_Text,
        text_style={},
    )
    STYLE.set_constraint("text_class", lambda val: issubclass(val, Button_Text))

    text = property(lambda self: self.text_widget.text)

    def __init__(self, parent, text=None, **kwargs):

        AbstractButton.__init__(self, parent, **kwargs)

        if text is not None:
            assert isinstance(text, str)
            self.text_widget = self.style["text_class"](self, text=text, layer=self.content, **self.style["text_style"])
            if self.name == "NoName":
                self._name = text




import time

from baopig.communicative import Communicative
from .utilities import present_time


class Stopwatch(Communicative):
    """
    A Stopwatch has the following signals :
        - START : emitted when the stopwatch starts
        - STOP : emitted when the stopwatch stops
        - RESET : emmitted when the stopwatch is reset
    """

    def __init__(self):

        Communicative.__init__(self)

        self._start_time = None
        self._stop_time = None

        self.create_signal("START")
        self.create_signal("STOP")
        self.create_signal("RESET")

    def __str__(self):

        return present_time(self.get_time())

    is_running = property(lambda self: self in _running_stopwatches)

    def get_time(self):
        """
        Return the current stopwatch time
        :return:
        """
        if self._start_time is None:
            return 0
        if self._stop_time is None:
            return time.time() - self._start_time
        return self._stop_time - self._start_time

    def reset(self):
        """
        Resets the stopwatch
        Only if the stopwatch is stopped
        """
        if self.is_running:
            raise PermissionError("You must stop a stopwatch before reset")
        if self.get_time() == 0:
            raise PermissionError("The stopwatch is already reset")

        self._start_time = None
        self._stop_time = None
        self.signal.RESET.emit()

    def start(self):
        """
        Starts the stopwatch
        A stopwatch can only be started if stopped
        """
        if _running_stopwatches.manager is None:
            raise PermissionError("Must launch an Application before starting a Stopwatch")
        if self.is_running:
            raise PermissionError("The stopwatch has already stared")

        if self._start_time is None:
            self._start_time = time.time()
        else:
            self._start_time = time.time() - self.get_time()
            self._stop_time = None
        _running_stopwatches.add(self)
        self.signal.START.emit()

    def stop(self):
        """
        Stops the stopwatch at current time
        """
        if not self.is_running:
            raise PermissionError("The stopwatch is not running")

        self._stop_time = time.time()
        _running_stopwatches.remove(self)
        self.signal.STOP.emit()


class _RunningStopwatches(set):
    pass


_running_stopwatches = _RunningStopwatches()
_running_stopwatches.manager = None





import time
import threading
from baopig.pybao.objectutilities import PrefilledFunction
from baopig.communicative import Communicative
from .utilities import *


class Timer(Communicative):
    """
    A Timer is an object who start at a given time called interval and then count down
    Once this time countdown is over, it emits the TIMEOUT signal
    You can connect commands to this signal trough the 'commands' parameter in constructor

    A Time has the following signals :
        - START : emitted when the timer starts
        - TIMEOUT : emitted when the countdown is over
    """
    def __init__(self, interval, command=lambda: None, *args, **kwargs):

        assert interval >= 0

        Communicative.__init__(self)

        self._interval = interval
        self._start_time = None
        self._pause_time = None
        self._end_time = None

        self.create_signal("TIMEOUT")
        self.create_signal("START")
        if args or kwargs:
            command = PrefilledFunction(command, *args, **kwargs)
        self.signal.TIMEOUT.connect(command, owner=None)

    interval = property(lambda self: self._interval)
    is_paused = property(lambda self: self._pause_time is not None)  # NOTE : a paused Timer is not running
    is_running = property(lambda self: self in _running_timers)

    def __repr__(self):

        return "Timer(interval={}, time_left={})".format(self.interval, self.get_time_left())

    def __str__(self):

        return present_time(self.get_time_left() if self.is_running else self.interval)

    def cancel(self):
        """
        Cancel the timer
        
        NOTE : It is preferable not to mess with emitting the TIMEOUT signal here.
               You can do it yourself if you know what you are doing.
        """

        if self._start_time is None:
            return

        with timer_lock:
            if self.is_running:
                _running_timers.remove(self)
            self._start_time = None
            self._pause_time = None
            self._end_time = None

    def get_time_left(self):

        if self._start_time is None:
            return 0

        if self.is_paused:
            return self.interval - (self._pause_time - self._start_time)

        return max(self._end_time - time.time(), 0)

    def pause(self):

        if not self.is_running:
            return

        with timer_lock:
            self._pause_time = time.time()
            _running_timers.remove(self)

    def resume(self):

        if self._pause_time is None:
            return

        with timer_lock:
            self._start_time = self._start_time + (time.time() - self._pause_time)
            self._end_time = self._start_time + self.interval
            self._pause_time = None
            _running_timers.add(self)

    def set_interval(self, interval):

        self._interval = interval

    def start(self):

        if self in _running_timers:
            raise PermissionError("The timer has already started")
        if _running_timers.manager is None:
            raise PermissionError("Must launch an Application before starting a Timer")

        with timer_lock:
            self._start_time = time.time()
            self._pause_time = None
            self._end_time = self._start_time + self.interval
            _running_timers.add(self)
            self.signal.START.emit()


class RepeatingTimer(Timer):
    """
    The repeat is carried by the manager

    If you enter a tuple for 'interval' parameter, the fisrt value will be used
    for the first interval, the second one for the other intervals
    """
    def __init__(self, interval, command, *args, **kwargs):

        if isinstance(interval, (int, float)):
            first_interval = repeat_interval = interval
        else:
            first_interval = interval[0]
            repeat_interval = interval[1]

        Timer.__init__(self, first_interval, command, *args, **kwargs)

        # interval returns the start of the current countdown
        self._first_interval = first_interval
        self._repeat_interval = repeat_interval

        self.create_signal("RESTART")

    first_interval = property(lambda self: self._first_interval)
    repeat_interval = property(lambda self: self._repeat_interval)

    def _repeat(self):
        """Should only be called by the manager"""

        left = self.get_time_left()
        assert left == 0, left
        assert self._start_time is not None

        with timer_lock:
            self._interval = self._repeat_interval
            self._start_time = self._end_time
            self._end_time += self._interval
            self.signal.RESTART.emit()

    def set_interval(self, interval):

        if isinstance(interval, (int, float)):
            self._first_interval = self._repeat_interval = self._interval = interval
        else:

            if self._interval is self._first_interval:
                self._interval = interval[0]
            else:
                self._interval = interval[1]
            self._first_interval = interval[0]
            self._repeat_interval = interval[1]


class SetattrTimer(Timer):
    """Set the attribute to value after a given delay (still a Timer, must be started)"""
    def __init__(self, owner, attr, val, delay):
        assert hasattr(owner, attr)
        Timer.__init__(self, delay, setattr, owner, attr, val)


class _RunningTimers(set):
    pass


timer_lock = threading.RLock()
_running_timers = _RunningTimers()
_running_timers.manager = None



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import time

from baopig.pybao import WeakList
from baopig.documentation import ApplicationExit
from baopig.io import keyboard, mouse, LOGGER
from .style import HasStyle, Theme, StyleClass
from .widget import Widget
from .utilities import *


class Application(HasStyle):
    """
    This is the main class in baopig
    It needs to be instanced before everything else
    """
    STYLE = StyleClass()

    def __init__(self, name=None, theme=None, size=None, mode=pygame.RESIZABLE):

        if name is None:
            name = self.__class__.__name__
        if theme is None:
            theme = Theme()

        HasStyle.__init__(self, theme)

        pygame.init()
        info = pygame.display.Info()

        self._name = name
        self._is_launched = False
        self._is_running = False
        self._fps = None
        self._default_mode = mode
        self._default_size = pygame.display.list_modes()[2] if size is None else size
        self._current_mode = self._current_size = None
        self._max_resolution = (info.current_w, info.current_h)
        self._is_fullscreen_TO_REMOVE = False
        self._scenes = WeakList()
        self._focused_scene = None
        self._caption = self.name
        self._painter = None  # To be set in self.launch()
        self._time_manager = None  # To be set in self.launch()

        # debug attributes
        self._debug_averagefps = False
        self._debug_launchtime = False

        mouse._application = self
        keyboard._application = self

        pygame.display.set_caption(self.name)

        self.launch_time = time.time()

    default_mode = property(lambda self: self._default_mode)
    default_size = property(lambda self: self._default_size)
    focused_scene = property(lambda self: self._focused_scene)
    fps = property(lambda self: self._fps)
    is_fullscreen = property(lambda self: bool(self.default_mode & pygame.FULLSCREEN))
    is_launched = property(lambda self: self._is_launched)
    max_resolution = property(lambda self: self._max_resolution)
    name = property(lambda self: self._name)
    painter = property(lambda self: self._painter)
    scenes = property(lambda self: self._scenes)
    size = property(lambda self: self._focused_scene.rect.size)

    def _add_scene(self, scene):

        assert scene not in self.scenes
        if not self.scenes:
            self._focused_scene = scene
        self.scenes.append(scene)

    def _manage_events(self):

        # TODO : solve : mouse entering and leaving the display are not properly handled
        #                      -> hovered widget may stay hovered

        # Events listening
        # Only apply on keyboard, mouse and application's operations
        events = pygame.event.get()
        for event in events:
            if event.type == pygame.KEYDOWN and event.key == pygame.K_F6:
                self.exit("FORCED EXIT (F6)")
            elif event.type == pygame.QUIT:
                self.exit()

            elif event.type in (pygame.MOUSEMOTION, pygame.MOUSEBUTTONDOWN, pygame.MOUSEBUTTONUP):
                mouse.receive(event)
            elif event.type in (pygame.KEYDOWN, pygame.KEYUP):
                keyboard.receive(event)
            elif event.type == pygame.ACTIVEEVENT:
                """
                gain  : 1 -> mouse in window
                        2 -> mouse outside window
                        
                state : 0 -> no state change
                        1 -> focus state just changed
                        2 -> iconify state just changed
                """
                if not hasattr(event, "gain"):  # Empty ACTIVEEVENT
                    continue
                if event.state == 2 and event.gain == 1:
                    self.refresh()
            elif event.type == pygame.VIDEORESIZE:
                if event.size != self.focused_scene.size:
                    self._default_size = event.size
                    self.focused_scene.resize(*event.size)
                else:
                    # We need an update
                    self.refresh()

            # DEFAULT SHORTKEYS
            if event.type == pygame.KEYDOWN:
                if keyboard.mod.ctrl:
                    # Cmd + e -> toggle debugging
                    if event.key == pygame.K_e:
                        if keyboard.mod.maj:
                            _ = self.focused_scene.children
                            raise Exception("Made for debugging")
                        self.toggle_debugging()
                    # Cmd + g -> collect garbage
                    elif event.key == pygame.K_g:
                        import gc
                        gc.collect()
                        LOGGER.info("Garbage collected")
                    # Cmd + r -> toggle recording (if Maj: save application.surface only when it changes)
                    elif event.key == pygame.K_r:
                        if self.painter.is_recording:
                            self.painter.stop_recording()
                        else:
                            self.painter.start_recording(only_at_change=keyboard.mod.maj)

                elif keyboard.mod.alt:
                    pass  # Nothing implemented

                elif event.key == pygame.K_ESCAPE:  # quit fullscreen or exit
                    if self.is_fullscreen:
                        self.set_display_mode(self.default_mode - pygame.FULLSCREEN)
                    else:
                        self.exit("pressed ESCAPE")
                # elif event.key == pygame.K_F5:  # fullscreen
                #     self.focused_scene.toggle_fullscreen()
                elif event.key == pygame.K_F4:  # minimize
                    self.iconify()
                elif event.key == pygame.K_F3:  # refresh
                    self.refresh()
                    LOGGER.info("Display refreshed")
                elif event.key == pygame.K_F2:  # screenshot TODO : A faire avec un clic droit
                    self.painter.screenshot()
                    LOGGER.info("Screenchot !")

            # TRANSMITTING EVENT
            focused = self.focused_scene.focused_widget
            if focused is not None:
                if event.type in (pygame.KEYDOWN, pygame.KEYUP):
                    assert focused.is_focused
                    assert focused.is_alive
                    assert focused.is_visible  # ?
                    assert focused.is_awake
                if event.type == pygame.KEYDOWN:
                    focused.handle_keydown(event.key)
                elif event.type == pygame.KEYUP:
                    focused.handle_keyup(event.key)
            # TODO
            # else:
            #     raise AssertionError

            # Events optionnal treatment
            self.focused_scene.handle_event(event)

    def _run(self):
        """
        Launch the application
        """

        try:

            self._is_running = True

            while self._is_running:

                # User events
                self._manage_events()

                # Possible advanced events treatments
                self._time_manager.update()
                self.focused_scene._container_run()

                # Possible coded stuff
                self.focused_scene.run()

                # If needed, drawing display
                if pygame.display.get_active():
                    self.painter._can_draw.set()

        except Exception as e:
            if isinstance(e, ApplicationExit):
                if str(e) != 'None':
                    LOGGER.info("{} exit : '{}'".format(self.__class__.__name__, e))
            else:
                raise e

        if self._debug_averagefps:
            fps_history = self.painter.fps_history
            if len(fps_history) > 0:
                fps_moy = sum(fps_history) / len(fps_history)
                LOGGER.info("{} ran with a global average of {} FPS"
                            "".format(self.__class__.__name__, fps_moy))
            else:
                LOGGER.info("{} didn't ran enough for a global FPS average".format(self.__class__.__name__))

        self._is_launched = False
        pygame.quit()

    def _update_display(self):
        """Updates display mode and size"""

        if self.is_launched:

            # self._is_fullscreen_TO_REMOVE = self.focused_scene.size == self.max_resolution

            # current_size = pygame.display.get_window_size()
            # current_mode = pygame.display.get

            size = self.focused_scene.asked_size
            if size is None:
                size = self.default_size

            # TODO : remove scene.mode ?
            mode = self.default_mode

            if size == self._current_size and mode == self._current_mode:
                Widget.set_surface(self.focused_scene, pygame.display.get_surface())
                return

            Widget.set_surface(self.focused_scene, pygame.display.set_mode(size, mode))

            if size != self._current_size:
                LOGGER.fine(f"Display size set to {size}")

            self._current_mode = mode
            self._current_size = size

    def set_debug(self, **kwargs):
        """
        arguments must be True or False
        keywords must be from this list :
            averagefps, launchtime
        """

        for kw, arg in kwargs.items():
            if arg is not None:
                assert hasattr(self, "_debug_" + kw)
                self.__setattr__("_debug_" + kw, arg)

    def exit(self, reason=None):

        self._is_running = False
        self.painter.stop()
        self.handle_app_close()

        # NOTE : after some tests, a pygame error makes it impossible to restart a closed application
        #        If I can make it work, and the program launches the app again
        # self._is_launched = False
        # self._current_size = self._current_mode = None

        raise ApplicationExit(reason)

    def flash(self):
        """
        Create a short flash on the application
        Can be used when an error occurs, like a wrong user input
        """
        pass  # TODO : implemented this functionnality

    def handle_app_close(self):
        """Stuff to do when the app is closed"""

    @staticmethod
    def iconify():

        with paint_lock:
            # if self.is_fullscreen:
            #     self.exit_fullscreen()  # TODO : test it
            pygame.display.iconify()

    def launch(self):

        if len(self.scenes) == 0:
            from baopig.prefabs.presentationscene import PresentationScene
            PresentationScene(self)

        assert not self.is_launched and pygame.get_init(), \
            "An application can only be launched once, due to pygame restrictions"
        self._is_launched = True

        from baopig.time.timemanager import time_manager
        self._time_manager = time_manager
        from baopig.threads import PainterThread
        self._painter = PainterThread(self)
        self.painter.set_fps(self._fps)

        assert self.focused_scene is not None
        scene = self.focused_scene
        self._focused_scene = None  # whitout this line, scene.open() does nothing, because it thinks it's already open
        scene.open()
        assert self.focused_scene is scene

        pygame.display.set_mode(self.size, self.default_mode)  # prevents a threading lag with painter.start()
        self.focused_scene.focus(self.focused_scene)
        self.painter.start()

        pygame.scrap.init()  # clipboard uses
        pygame.event.get()  # ignore events that took place during the app's load
        mouse._pos = pygame.mouse.get_pos()
        self._run()

    def open(self, scene):

        if isinstance(scene, str):
            for s in self.scenes:
                if s.name == scene:
                    scene = s
                    break
        if scene not in self.scenes:
            raise PermissionError(f"Unknown scene : '{scene}' "
                                  f"(existing scenes : {tuple(str(s) for s in self.scenes)})")

        scene.open()

    def refresh(self):
        """
        Send a paintrequest to every container in focused_scene
        if only_containers is False, sends a paintrequest to every focused_scene's children
        """
        # TODO : refresh & screenshot buttons by default
        self.focused_scene._container_refresh(recursive=True, only_containers=False)

    def set_caption(self, title, icontitle=""):

        self._caption = title
        pygame.display.set_caption(title, icontitle)

    def set_default_size(self, size):  # TODO : resize ?

        if size == self.default_size:
            return
        self._default_size = size
        for scene in self.scenes:
            if scene.asked_size is not None:
                scene.resize(*self.default_size)

    def set_fps(self, fps):

        self._fps = fps
        if self.painter is not None:
            self.painter.set_fps(fps)

    @staticmethod
    def set_icon(icon):

        pygame.display.set_icon(icon)

    def set_display_mode(self, mode):

        if mode is self.default_mode:
            return

        assert mode in (0, pygame.NOFRAME, pygame.RESIZABLE, pygame.FULLSCREEN)

        # if mode is pygame.FULLSCREEN and self.mode != mode:
        #   self._mode_before_fullscreen = self.focused_scene.mode
        # self._size_before_fullscreen = self.asked_size

        # mode = 0

        self._default_mode = mode
        self._update_display()

    def toggle_debugging(self):

        self.focused_scene.toggle_debugging()

    def warning(self, i_dont_know):

        # TODO : show the warning on screen
        raise PermissionError("Not implemented")



from baopig.pybao.objectutilities import Object, deque
from baopig.io import keyboard, mouse, LOGGER
from baopig.lib import *
from .scrollview import ScrollView
from .text import Text


# TODO : LineEntry (with select_all_on_focus and exec_on_defocus)
# TODO : presentation text when nothing is in the text ?


class SelectionRect_For_TextEdit(SelectionRect):

    def clip(self, rect):
        pass


class TextEdit(ScrollView, Selector):
    STYLE = ScrollView.STYLE.substyle()
    STYLE.modify(
        background_color="theme-color-font_opposite",
        height=45,
        padding=2,
        width=100,
    )

    def __init__(self, parent, text="", **kwargs):

        text_kwargs = {}
        for key in (
                "font_file",
                "font_height",
                "font_color",
                "font_bold",
                "font_italic",
                "font_underline",
                "max_width",
        ):
            if key in kwargs:
                text_kwargs[key] = kwargs.pop(key)

        Selector.__init__(self, parent, can_select=True, selectionrect_class=SelectionRect_For_TextEdit, **kwargs)
        ScrollView.__init__(self, parent)

        self._text_widget_ref = Text(self, text=text, selectable=True, **text_kwargs).get_weakref()
        # self.text_widget.signal.MOTION.disconnect(self.text_widget.handle_unselect)

        self.set_selectionrect_visibility(False)

        self._cursor_ref = Cursor(self).get_weakref()
        self.cursor.sleep()

        self.set_style_for(SelectionRect, ref=self.text_widget)

    cursor = property(lambda self: self._cursor_ref())
    text = property(lambda self: self._text_widget_ref().get_text())
    text_widget = property(lambda self: self._text_widget_ref())

    def accept(self, text):
        return text != ''

    def del_selection_data(self):

        if not self.text_widget.is_selected:
            return
        cursor_index = self.text_widget.find_index(
            char_index=self.text_widget.line_selections[0].index_start,
            line_index=self.text_widget.line_selections[0].line.line_index
        )
        assert self.is_selecting
        selected_lines = tuple(line for line in self.text_widget.lines if line.is_selected)
        if selected_lines:
            if self.cursor is not None:
                self.cursor.save()
            for line in selected_lines:
                line._text = line.text[:line.selection.index_start] + line.text[line.selection.index_end:]
                line._end = '' if line.selection._is_selecting_line_end else line.end
            line.config(line.text)
        self.close_selection()
        self.cursor.config(text_index=cursor_index)

    def end_selection(self, *args, **kwargs):

        super().end_selection(*args, **kwargs)

        pos = (self.selection_rect.abs_end[0] - self.text_widget.abs_rect.left,
               self.selection_rect.abs_end[1] - self.text_widget.abs_rect.top)
        line_index, char_index = self.text_widget._find_indexes(pos=pos)
        if line_index != self.cursor.line_index or char_index != self.cursor.char_index:
            self.cursor.config(line_index=line_index, char_index=char_index, selecting="done")

    def handle_defocus(self):

        super().handle_defocus()
        self.cursor.sleep()

    def handle_focus(self):

        super().handle_focus()

        text_index = self.text_widget._find_index(  # stupid looking bug fix
            pos=(self.x_scroller.val + self.rect.width, self.y_scroller.val + self.rect.height)
        )

        self.cursor.wake()
        self.cursor.config(text_index=text_index)

    def handle_keydown(self, key):

        if keyboard.mod.ctrl:
            if key in (pygame.K_a, pygame.K_c, pygame.K_v, pygame.K_x):
                return super().handle_keydown(key)
        self.cursor.handle_keydown(key)

    def handle_link(self):

        super().handle_link()

        if not mouse.has_double_clicked and not mouse.has_triple_clicked:  # else, the cursor follow the selection
            self.cursor.wake()
            self.cursor.config(text_index=self.text_widget._find_mouse_index())

    def paste(self, data):

        self.cursor.write(data)

    def select_all(self):

        if self.selectables:
            if self.is_selecting:
                self.close_selection()
            self.start_selection(self.main_widget.abs_rect.topleft)
            self.end_selection(self.main_widget.abs_rect.bottomright)

    def set_text(self, text):

        self.text_widget.set_text(text)


class Cursor(Rectangle, RepetivelyAnimated):
    """
    By default, at creation, a cursor is set at mouse position
    """
    STYLE = Rectangle.STYLE.substyle()
    STYLE.modify(
        color="theme-color-font"
    )

    def __init__(self, parent):

        assert isinstance(parent, TextEdit)

        h = parent.text_widget.font.height

        Rectangle.__init__(
            self,
            parent=parent,
            ref=parent.text_widget,
            size=(int(h / 10), h),
            name=parent.name + " -> cursor",
            layer=Layer(parent, Cursor, name="cursor_layer", level=LayersManager.FOREGROUND, maxlen=1),
            touchable=False
        )
        RepetivelyAnimated.__init__(self, parent, interval=.5)

        self._char_index = 0  # index of cursor position, see _Line._chars_pos for more explanations
        self.__line_index = 0  # index of cursor line, see Text._lines_pos for more explanations
        self._line = None
        self._text_index = 0  # index of cusor in Text.text

        # History
        """
        A History element is created when :
            - A new text insert
            - A part of text pop
            - Just before a selected data is delete

        A History element store these data :
            - the entire text of parent
            - the cusror indexes (line and char)
            - the selection start and end, if the parent was selecting
        """
        max_item_stored = 50
        self.history = deque(maxlen=max_item_stored)
        self.back_history = deque(maxlen=max_item_stored)

    char_index = property(lambda self: self._char_index)

    def _set_line_index(self, li):
        self.__line_index = li
        self._line = self.text_widget.lines[li]

    _line_index = property(lambda self: self.__line_index, _set_line_index)
    left_from_text = property(lambda self: self.rect.left - self.text_widget.rect.left)
    line_index = property(lambda self: self.__line_index)
    line = property(lambda self: self._line)
    text_widget = property(lambda self: self._parent.text_widget)
    text_index = property(lambda self: self._text_index)

    def clip(self):
        """ Scroll the Text until the cursor is inside the TextEdit"""

        x_scroller = self.parent.x_scroller
        if self.rect.right > self.parent.rect.width:
            dx = self.rect.right - self.parent.rect.width
            new_val = min(x_scroller.val + dx + 30, x_scroller.maxval)
            x_scroller.set_val(new_val)
        elif self.rect.left < 0:
            dx = self.rect.left
            new_val = max(x_scroller.val + dx - 30, x_scroller.minval)
            x_scroller.set_val(new_val)

        y_scroller = self.parent.y_scroller
        if self.rect.bottom > self.parent.rect.height:
            dy = self.rect.bottom - self.parent.rect.height
            new_val = min(y_scroller.val + dy, y_scroller.maxval)
            y_scroller.set_val(new_val)
        elif self.rect.top < 0:
            dy = self.rect.top
            new_val = max(y_scroller.val + dy, y_scroller.minval)
            y_scroller.set_val(new_val)

    def config(self, text_index=None, line_index=None, char_index=None, selecting=False, save=True):
        """
        Place the cursor at line n° line_index and before the character n° char_index, count from 0
        If text_index is given instead of line_index and char_index, we use parent.find_indexes

        If char_index is at the end of a cutted line (a line too big for the text width), then
        the cursor can either be on the end of the line or at the start of the next line, it is
        algorithmically the same. So the object who config the cursor will decide where to place the
        cursor. It can give a float value for text_index (like 5.4) wich mean "Hey, if the cursor is
        at the end of a cutted line, let it move the start of the next one." In this exemple, the
        text_index value will be 5. This works also with char_index = 5.4
        """

        if text_index is not None:
            assert line_index is None
            assert char_index is None
            line_index, char_index = self.text_widget.find_indexes(text_index=text_index)
        else:
            assert char_index is not None
            assert line_index is not None
            text_index = self.text_widget.find_index(line_index, char_index)

        # assert text_index == self.text_widget.find_index(line_index, char_index)

        if selecting:
            if self.parent.selection_rect is None:
                abs_pos = self.abs_rect.center
                self.parent.start_selection(abs_pos)

        def fit(v, mini, maxi):
            if v < mini:
                v = mini
            elif v > maxi:
                v = maxi
            return v

        self._text_index = fit(text_index, 0, len(self.text_widget.text))
        self._line_index = fit(line_index, 0, len(self.text_widget.lines) - 1)
        self._char_index = fit(char_index, 0, len(self.line.text))

        if self.char_index == len(self.line.text_with_end):
            LOGGER.warning("Tricky cursor position")

        if self.get_weakref()._ref is None:
            LOGGER.warning('This widget should be dead :', self)

        old_pos = self.rect.pos
        self.set_pos(top=self.line.rect.top + self.text_widget.rect.top)
        self.set_pos(left=self.line.find_pixel(self.char_index) + self.text_widget.rect.left)

        self.clip()

        self.start_animation()
        self.show()  # always shown when the animation starts

        if selecting == "done":
            pass
        elif selecting is True:
            if self.parent.selection_rect.abs_end is None or old_pos != self.rect.pos:
                self.parent.end_selection(self.abs_rect.topleft)
        elif selecting is False:
            if self.parent.is_selecting:
                self.parent.close_selection()
        else:
            raise PermissionError

        if save and (not self.history or self.text_widget.text != self.history[-1].text):
            self.save()

    def handle_keydown(self, key):
        """
        N'accepte que les evenements du clavier
        Si la touche est speciale, effectue sa propre fonction
        Modifie le placement du curseur
        """

        # Cmd + ...
        if keyboard.mod.ctrl:
            # Maj + Cmd + ...
            if keyboard.mod.maj:
                if key == pygame.K_z:
                    self.redo()
                return
            elif keyboard.mod.cmd or keyboard.mod.alt:
                return
            elif key == pygame.K_d:
                # Duplicate
                selected_data = self.parent.get_selection_data()
                if selected_data == '':
                    selected_data = self.line.text_with_end
                    self.line.insert(0, selected_data)
                else:
                    self.parent.close_selection()
                    self.line.insert(self.char_index, selected_data)
                self.config(text_index=self.text_index + len(selected_data))
            elif False and key == pygame.K_r:  # TODO
                # Execute
                try:
                    exec(self.text_widget.text)
                except Exception as e:
                    LOGGER.warning("CommandError: " + str(e))
            elif key == pygame.K_z:
                self.undo()
            elif key in (pygame.K_LEFT, pygame.K_HOME):
                self.config(self.text_widget.find_index(line_index=self.line_index, char_index=0),
                            selecting=keyboard.mod.maj)
            elif key in (pygame.K_RIGHT, pygame.K_END):
                self.config(self.text_widget.find_index(line_index=self.line_index, char_index=len(self.line.text)),
                            selecting=keyboard.mod.maj)
            elif key == pygame.K_UP:
                if self.line_index > 0:
                    self.config(line_index=0,
                                char_index=self.text_widget.lines[0].find_index(self.left_from_text),
                                selecting=keyboard.mod.maj)
            elif key == pygame.K_DOWN:
                if self.line_index < len(self.text_widget.lines) - 1:
                    self.config(
                        line_index=len(self.text_widget.lines) - 1,
                        char_index=self.text_widget.lines[len(self.text_widget.lines) - 1].find_index(
                            self.left_from_text),
                        selecting=keyboard.mod.maj
                    )

        # Cursor movement
        elif 1073741898 <= key <= 1073741906 and key != 1073741900:

            if key in (pygame.K_LEFT, pygame.K_RIGHT):

                if keyboard.mod.alt:  # go to word side
                    if key == pygame.K_LEFT:
                        if self.char_index == 0:
                            return
                        self.config(text_index=self.text_index - 1, selecting=keyboard.mod.maj)
                        while self.char_index > 0 and \
                                (self.line.text[self.char_index - 1] != ' ' or self.line.text[self.char_index] == ' '):
                            self.config(text_index=self.text_index - 1, selecting=keyboard.mod.maj)
                    elif key == pygame.K_RIGHT:
                        if self.char_index == len(self.line.text):
                            return
                        self.config(text_index=self.text_index + 1, selecting=keyboard.mod.maj)
                        while self.char_index < len(self.line.text) and \
                                (self.line.text[self.char_index - 1] != ' ' or self.line.text[self.char_index] == ' '):
                            self.config(text_index=self.text_index + 1, selecting=keyboard.mod.maj)
                elif (not keyboard.mod.maj) and len(self.text_widget.line_selections):
                    if key == pygame.K_LEFT:
                        self.config(line_index=self.text_widget.line_selections[0].line_index,
                                    char_index=self.text_widget.line_selections[0].index_start)
                    elif key == pygame.K_RIGHT:
                        self.config(line_index=self.text_widget.line_selections[-1].line_index,
                                    char_index=self.text_widget.line_selections[-1].index_end)
                elif key == pygame.K_LEFT:
                    # self.config(char_index=self.char_index-1, selecting=keyboard.mod.maj)
                    self.config(text_index=self.text_index - 1, selecting=keyboard.mod.maj)
                elif key == pygame.K_RIGHT:
                    # self.config(char_index=self.char_index+1, selecting=keyboard.mod.maj)
                    self.config(text_index=self.text_index + 1, selecting=keyboard.mod.maj)

            elif key in (pygame.K_HOME, pygame.K_END):
                if key == pygame.K_HOME:  # Fn + K_LEFT
                    self.config(text_index=self.text_widget.find_index(line_index=self.line_index, char_index=0),
                                selecting=keyboard.mod.maj)
                elif key == pygame.K_END:  # Fn + K_RIGHT
                    self.config(self.text_widget.find_index(line_index=self.line_index, char_index=len(self.line.text)),
                                selecting=keyboard.mod.maj)

            elif key in (pygame.K_UP, pygame.K_DOWN):
                if key == pygame.K_UP:
                    if self.line_index > 0:
                        self.config(line_index=self.line_index - 1,
                                    char_index=self.text_widget.lines[self.line_index - 1].find_index(
                                        self.left_from_text),
                                    selecting=keyboard.mod.maj)
                if key == pygame.K_DOWN:
                    if self.line_index < len(self.text_widget.lines) - 1:
                        self.config(line_index=self.line_index + 1,
                                    char_index=self.text_widget.lines[self.line_index + 1].find_index(
                                        self.left_from_text),
                                    selecting=keyboard.mod.maj)

            elif key in (pygame.K_PAGEUP, pygame.K_PAGEDOWN):
                if key == pygame.K_PAGEUP:
                    if self.line_index > 0:
                        self.config(line_index=0,
                                    char_index=self.text_widget.lines[0].find_index(self.left_from_text),
                                    selecting=keyboard.mod.maj)
                if key == pygame.K_PAGEDOWN:
                    if self.line_index < len(self.text_widget.lines) - 1:
                        self.config(
                            line_index=len(self.text_widget.lines) - 1,
                            char_index=self.text_widget.lines[len(self.text_widget.lines) - 1].find_index(
                                self.left_from_text),
                            selecting=keyboard.mod.maj)

        # Suppression
        elif key == pygame.K_BACKSPACE:
            if self.text_widget.is_selected:
                self.parent.del_selection_data()
            elif self.line_index > 0 or self.char_index > 0:
                if self.char_index > 0:
                    self.line.pop(self.char_index-1)
                else:
                    self.text_widget.lines[self.line_index - 1].pop(-1)
                old = self.text_index
                self.config(text_index=self.text_index - 1)
                assert self.text_index == old - 1

        elif key == pygame.K_DELETE:
            if self.parent.is_selected:
                self.text_widget.del_selection_data()
            if self.line.end == '' and self.char_index == len(self.line.text):
                if self.line_index < len(self.text_widget.lines) - 1:
                    self.text_widget.lines[self.line_index + 1].pop(0)
            else:
                self.line.pop(self.char_index)

            # We don't use text_index because, if self.char_index is 0,
            # we want to stay at 0, text_index might send the cursor at the
            # end of the previous line if it is a cutted line
            self.config(line_index=self.line_index,
                        char_index=self.char_index)

        elif key == pygame.K_ESCAPE:
            self.parent.defocus()

        # Write
        else:
            unicode = keyboard.last_event.unicode
            if key == pygame.K_RETURN:
                unicode = '\n'
            elif key == pygame.K_TAB:
                unicode = '    '
            if unicode:  # skip every maj, majlock, met, cmd, ctrl...
                self.write(unicode)

    def write(self, string):

        if self.text_widget.is_selected:
            self.parent.del_selection_data()

        # Letters (lowercase and uppercase)
        text = self.text_widget.text[:self.char_index] + string + self.text_widget.text[self.char_index:]
        if self.parent.accept(text):
            self.line.insert(self.char_index, string)
            self.config(text_index=self.text_index + len(string))
            # TODO : solve : when a TextEdit is resized, the cursor does not follow its text

    # History
    def redo(self):
        """
        Restaure la derniere modification
        """
        if self.back_history:

            backup = self.back_history.pop()  # last element of self.back_history, the current state
            self.history.append(backup)

            backup = self.history[-1]
            self.parent.set_text(backup.text)
            self.config(line_index=backup.cursor_line_index, char_index=backup.cursor_char_index, save=False)
            if backup.selection_start is not None:
                if self.parent.is_selecting:
                    self.parent.close_selection()
                self.parent.start_selection(backup.selection_start)
                self.parent.end_selection(backup.selection_end)

        # else:
        #     LOGGER.info("Cannot redo last operation because the operations history is empty")

    def save(self):

        # if self.parent.is_selecting:
        current = Object(
            text=self.text_widget.text,
            cursor_line_index=self.line_index,
            cursor_char_index=self.char_index,
            selection_start=self.parent.selection_rect.abs_start if self.parent.selection_rect else None,
            selection_end=self.parent.selection_rect.abs_end if self.parent.selection_rect else None
        )
        self.history.append(current)
        self.back_history.clear()

    def undo(self):
        """
        Annule la derniere modification
        """
        if len(self.history) > 1:  # need at least 2 elements in history

            backup = self.history.pop()  # last element of self.history, which is the state before undo()
            self.back_history.append(backup)

            previous = self.history[-1]
            self.parent.set_text(previous.text)
            self.config(line_index=previous.cursor_line_index, char_index=previous.cursor_char_index, save=False)
            if previous.selection_start is not None:
                if self.parent.is_selecting:
                    self.parent.close_selection()
                self.parent.start_selection(previous.selection_start)
                self.parent.end_selection(previous.selection_end)
        # else:
        #     LOGGER.info("Cannot undo last operation because the operations history is empty")
